/*
	mod_v2h
	Michael Link 2000/2001 [04152002]
	mlink@fractal.net
*/

#undef AP_HAVE_DESIGNATED_INITIALIZER

#include "apr.h"
#include "apr_strings.h"
#include "apr_lib.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
//#include "util_uri.h"
#include "apr_uri.h"

#include "ap_config_auto.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <mysql.h>

#define	mkupper(c)	(((c) >= 'a' && (c) <= 'z') ? ((c) - 'a' + 'A') : c)

int match(const char* pattern, const char* string, register char delim = ' ')
{
	char type;

	if (!pattern || !string) {
		return 0;
	}

	while (*string && *pattern && *pattern != '*' && *pattern != '%') {
		if (*pattern == '\\' && *(pattern + 1)) {
			if (!*++pattern || !(mkupper(*pattern) == mkupper(*string))) {
				return 0;
			}
			else {
				pattern++, string++;
			}
		}


		if (*pattern == '?') {
			pattern++, string++;
		}
		else if (mkupper(*pattern) == mkupper(*string)) {
			pattern++, string++;
		}
		else {
			break;
		}
	}
	
	if (*pattern == '*' || *pattern == '%') {
		type = (*pattern++);
		while (*string) {
			if (match(pattern, string)) {
				return 1;
			}
			else if (type == '*' || *string != delim) {
				string++;
			}
			else {
				break;
			}
		}
	}
	
	if (!*string && !*pattern) {
		return 1;
	}
	
	return 0;
}

inline void* operator new(size_t sz, apr_pool_t *pl)
{
	return apr_palloc(pl, sz);
}

inline void* operator new[](size_t sz, apr_pool_t *pl)
{
	return apr_palloc(pl, sz);
}

// want to call destructors without actually freeing memory from the pool
inline void operator delete(void* ptr) { }
inline void operator delete[](void* ptr) { }

typedef unsigned long osize_t;

template <class T>
class plist {
  protected:
	osize_t sz;

  public:
	class node {
	  public:
		T e;
		node *prev, *next;
		
		node(void)
		{
			prev = next = NULL;
		}
		
		~node(void)
		{
			// void
		}
	};

	apr_pool_t *p;
	node *top, *end;
	
	plist(apr_pool_t *pl) : p(pl)
	{
		sz = 0;
		top = end = NULL;
	}
	
	virtual ~plist(void)
	{
		zero();
	}
	
	virtual node* add(void)
	{
		node* n = new(p) node;
	
		if (!top) { // add first element
			top = end = n;
		}
		else { // append to end
			n->prev = end;
			end->next = n;
			end = n;
		}
		
		sz++;
		
		return n;
	}
	
	virtual void del(register node* d)
	{
		if ((d == top) && (d == end)) {
			top = end = NULL;
		}
		else if (d == top) {
			top = d->next;
			top->prev = NULL;
		}
		else if (d == end) {
			end = d->prev;
			end->next = NULL;
		}
		else {
			d->prev->next = d->next;
			d->next->prev = d->prev;
		}
		
		delete d;
	
		sz--;
	}
	
	void zero(void)
	{
		while (top) {
			del(top);
		}
	}
	
	osize_t size(void)
	{
		return sz;
	}
};

class plist<char*> {
  protected:
	osize_t sz;

  public:
	class node {
	  public:
		char *e;
		node *prev, *next;
		
		node(void)
		{
			prev = next = NULL;
		}
		
		~node(void)
		{
			delete [] e;
		}
	};

	apr_pool_t *p;
	node *top, *end;
	
	plist(apr_pool_t *pl) : p(pl)
	{
		sz = 0;
		top = end = NULL;
	}
	
	virtual ~plist(void)
	{
		zero();
	}
	
	virtual node* add(void)
	{
		node* n = new(p) node;
	
		if (!top) { // add first element
			top = end = n;
		}
		else { // append to end
			n->prev = end;
			end->next = n;
			end = n;
		}
		
		sz++;
		
		return n;
	}
	
	virtual void del(register node* d)
	{
		if ((d == top) && (d == end)) {
			top = end = NULL;
		}
		else if (d == top) {
			top = d->next;
			top->prev = NULL;
		}
		else if (d == end) {
			end = d->prev;
			end->next = NULL;
		}
		else {
			d->prev->next = d->next;
			d->next->prev = d->prev;
		}
		
		delete d;
	
		sz--;
	}
	
	void zero(void)
	{
		while (top) {
			del(top);
		}
	}
	
	osize_t size(void)
	{
		return sz;
	}
};

static int v2h_init_handler(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s);
static void* v2h_create_srv_config(apr_pool_t *p, server_rec *s);

static int v2h_translate(request_rec *r);
static int v2h_handler(request_rec *r);

static const char* sflg(cmd_parms *parms, void *mconfig, int flag);
static const char* sfld(cmd_parms *parms, void *mconfig, const char *arg);
static const char* suint(cmd_parms *parms, void *mconfig, const char *arg);
static const char* slst(cmd_parms *parms, void *mconfig, const char *arg);

enum {
	f_on	= 1,
	f_image	= 2
};

typedef const char* (*fz_t)(void);

#define VHE_CONFIG_SZ 10

class v2h_config_rec {
  public:
	int				fl;
	char			*id[VHE_CONFIG_SZ];
	unsigned int	port;
	plist<char*>	*decline_list;
	
	v2h_config_rec(apr_pool_t *p)
	{
		fl			= 0;
		port		= 0;
		
		id[0]		= NULL;							// mysql db
		id[1]		= NULL;							// mysql table
		id[2]		= NULL;							// mysql server field
		id[3]		= NULL;							// mysql path field
		id[4]		= NULL;							// mysql hostname
		id[5]		= NULL;							// mysql pass
		id[6]		= NULL;							// mysql user
		id[7]		= NULL;							// mysql env table
		id[8]		= NULL;							// path header
		id[9]		= NULL;							// image path
		
		decline_list	= new(p) plist<char*>(p);
	}
};

static const command_rec v2h_commands[] = {
	AP_INIT_FLAG(		"v2h",					(fz_t) sflg,		(void*) f_on,		RSRC_CONF,		"turn mod_v2 on or off." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Db",			(fz_t) sfld,		(void*) 0,			RSRC_CONF,		"set mysql database." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Tbl",		(fz_t) sfld,		(void*) 1,			RSRC_CONF,		"set mysql table." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Serv_Fld",	(fz_t) sfld,		(void*) 2,			RSRC_CONF,		"set mysql server field." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Path_Fld",	(fz_t) sfld,		(void*) 3,			RSRC_CONF,		"set mysql path field." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Host",		(fz_t) sfld,		(void*) 4,			RSRC_CONF,		"set mysql host." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Port",		(fz_t) suint,		(void*) 0,			RSRC_CONF,		"set mysql port." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Pass",		(fz_t) sfld,		(void*) 5,			RSRC_CONF,		"set mysql pass." ),
	AP_INIT_TAKE1(		"v2h_Mysql_User",		(fz_t) sfld,		(void*) 6,			RSRC_CONF,		"set mysql user." ),
	AP_INIT_TAKE1(		"v2h_Mysql_Env_Fld",	(fz_t) sfld,		(void*) 7,			RSRC_CONF,		"set mysql environment table." ),
	AP_INIT_TAKE1(		"v2h_PathHead",			(fz_t) sfld,		(void*) 8,			RSRC_CONF,		"set path header." ),
	AP_INIT_FLAG(		"v2h_UseImage",			(fz_t) sflg,		(void*) f_image,	RSRC_CONF,		"use memory image cache." ),
	AP_INIT_TAKE1(		"v2h_ImagePath",		(fz_t) sfld,		(void*) 9,			RSRC_CONF,		"image path." ),
	AP_INIT_ITERATE(	"v2h_DeclineURI",		(fz_t) slst,		NULL,				RSRC_CONF,		"declined path." ),
	{ NULL }
};

static void register_hooks(apr_pool_t *p)
{
	ap_hook_post_config(v2h_init_handler, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_translate_name(v2h_translate, NULL, NULL, APR_HOOK_MIDDLE);
	ap_hook_handler(v2h_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

AP_DECLARE_DATA module v2h_module = {
    STANDARD20_MODULE_STUFF,
    NULL,							/* create per-directory config structure */
    NULL,							/* merge per-directory config structures */
    v2h_create_srv_config,			/* create per-server config structure */
    NULL,							/* merge per-server config structures */
    v2h_commands,					/* command apr_table_t */
    register_hooks					/* register hooks */
};

static void* v2h_create_srv_config(apr_pool_t *p, server_rec *s)
{
	v2h_config_rec *v2hr = new(p) v2h_config_rec(p);
	
	return v2hr;
}

static const char* sflg(cmd_parms *parms, void *mconfig, int flag)
{
	int f = (int) parms->info;
	v2h_config_rec *v2hr = (v2h_config_rec*) ap_get_module_config(parms->server->module_config, &v2h_module);
	
	if (flag) {
		v2hr->fl |= f;
	}
	else {
		v2hr->fl &= ~f;
	}
	
	return NULL;
}

static const char* sfld(cmd_parms *parms, void *mconfig, const char *arg)
{
	int pos = (int) parms->info;
	v2h_config_rec *v2hr = (v2h_config_rec*) ap_get_module_config(parms->server->module_config, &v2h_module);
	
	v2hr->id[pos] = apr_pstrdup(parms->pool, arg);
	
	return NULL;
}

static const char* suint(cmd_parms *parms, void *mconfig, const char *arg)
{
	int pos = (int) parms->info;
	unsigned int ui = (unsigned int) atoi(arg);
	v2h_config_rec *v2hr = (v2h_config_rec*) ap_get_module_config(parms->server->module_config, &v2h_module);
	
	memcpy(&(v2hr->port), &ui, sizeof(int));
	
	return NULL;
}

static const char* slst(cmd_parms *parms, void *mconfig, const char *arg)
{
	v2h_config_rec *v2hr = (v2h_config_rec*) ap_get_module_config(parms->server->module_config, &v2h_module);
	plist<char*>::node *n = v2hr->decline_list->add();
	
	n->e = apr_pstrdup(parms->pool, arg);
	
	return NULL;
}

static char* mysql_host_check(request_rec *r, const char *host, char *&env, v2h_config_rec *v2hr)
{
	MYSQL *mysql = NULL;
	MYSQL_RES *res = NULL;
	MYSQL_ROW row;
	char *q, *rstr;
	int i;
	
	if (!(v2hr->fl & f_on)) {
		return NULL;
	}
	
	if (!(mysql = mysql_init(NULL))) {
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: mysql_host_check: unable to allocate MYSQL connection.");

		return NULL;
	}
	
	if (!mysql_real_connect(mysql, v2hr->id[4], v2hr->id[6], v2hr->id[5], v2hr->id[0], v2hr->port, NULL, CLIENT_COMPRESS)) {
		ap_log_error(APLOG_MARK, APLOG_ALERT, 0, r->server, "v2h: mysql_host_check: %s.", mysql_error(mysql));

		mysql_close(mysql);
		return NULL;
	}
	
	if (v2hr->id[7]) {
		q = apr_psprintf(r->pool, "SELECT %s,%s FROM %s WHERE %s = '%s'", v2hr->id[3], v2hr->id[7], v2hr->id[1], v2hr->id[2], host);
	}
	else {
		q = apr_psprintf(r->pool, "SELECT %s FROM %s WHERE %s = '%s'", v2hr->id[3], v2hr->id[1], v2hr->id[2], host);
	}
	
	if (mysql_query(mysql, q)) {
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: mysql_host_check: %s.", mysql_error(mysql));

		mysql_close(mysql);
		return NULL;
	}
	
	if (!(res = mysql_store_result(mysql))) {
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: mysql_host_check: %s.", mysql_error(mysql));

		mysql_close(mysql);
		return NULL;
	}
	
	switch (mysql_num_rows(res)) {
	  case 1:
		break;
	  case 0:
		mysql_free_result(res);
		mysql_close(mysql);
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: mysql_host_check: nothing found: %s", q);
		return NULL;
	  default:
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: mysql_host_check: %s has more than 1 server row.", host);

		mysql_free_result(res);
		mysql_close(mysql);
		return NULL;
	}
	
	if (!(row = mysql_fetch_row(res))) {
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: mysql_host_check: %s.", mysql_error(mysql));

		mysql_free_result(res);
		mysql_close(mysql);
		return NULL;
	}
	
	rstr = new(r->pool) char[strlen(row[0]) + 1];
	strcpy(rstr, row[0]);
	
	if (v2hr->id[7]) {
		env = new(r->pool) char[strlen(row[1]) + 1];
		strcpy(env, row[1]);
	}
	
	mysql_free_result(res);
	mysql_close(mysql);
	
	return rstr;
}

struct mem_cache {
	char	domain[256], path[64], extra[32];
};

static int mccmp(const void *e1, const void *e2)
{
	mem_cache *mc1 = (mem_cache*) e1, *mc2 = (mem_cache*) e2;
	
	return strcmp(mc1->domain, mc2->domain);
}

static char* path_lookup(request_rec *r, const char *host, char *&env, v2h_config_rec *v2hr)
{
	if (v2hr->fl & f_image) {
		struct stat st;
		
		if (stat(v2hr->id[9], &st) == -1) {
			MYSQL *mysql = NULL;
			MYSQL_RES *res = NULL;
			MYSQL_ROW row;
			mem_cache *mcptr, mckey, *mcval;
			size_t nel;
			int fd;
			char *q;
			
			if (!(mysql = mysql_init(NULL))) {
				ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: path_lookup (create img phase): unable to allocate MYSQL connection.");

				return NULL;
			}
			
			if (!mysql_real_connect(mysql, v2hr->id[4], v2hr->id[6], v2hr->id[5], v2hr->id[0], v2hr->port, NULL, CLIENT_COMPRESS)) {
				ap_log_error(APLOG_MARK, APLOG_ALERT, 0, r->server, "v2h: path_lookup (create img phase): %s.", mysql_error(mysql));

				mysql_close(mysql);
				return NULL;
			}
			
			if (v2hr->id[7]) {
				q = apr_psprintf(r->pool, "SELECT %s,%s,%s FROM %s", v2hr->id[2], v2hr->id[3], v2hr->id[7], v2hr->id[1]);
			}
			else {
				q = apr_psprintf(r->pool, "SELECT %s,%s FROM %s", v2hr->id[2], v2hr->id[3], v2hr->id[1]);
			}
			
			if (mysql_query(mysql, q)) {
				ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: path_lookup (create img phase): %s.", mysql_error(mysql));

				mysql_close(mysql);
				return NULL;
			}
			
			if (!(res = mysql_store_result(mysql))) {
				ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: path_lookup (create img phase): %s.", mysql_error(mysql));

				mysql_close(mysql);
				return NULL;
			}
			
			nel = mysql_num_rows(res);
			mcptr = (mem_cache*) calloc(nel, sizeof(mem_cache));
			
			for (int i = 0; row = mysql_fetch_row(res); i++) {
				strncpy(mcptr[i].domain, row[0], 255);
				mcptr[i].domain[255] = '\0';
				
				strncpy(mcptr[i].path, row[1], 63);
				mcptr[i].path[63] = '\0';
				
				if (v2hr->id[7]) {
					strncpy(mcptr[i].extra, row[2], 31);
					mcptr[i].extra[31] = '\0';
				}
			}
			
			qsort(mcptr, nel, sizeof(mem_cache), mccmp);
			
			if ((fd = open(v2hr->id[9], O_WRONLY|O_CREAT|O_TRUNC, 0644)) == -1) {
				ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: path_lookup (create img phase): unable to write to: %s: %s.", v2hr->id[9], strerror(errno));
				v2hr->fl &= ~f_image;
			}
			else {
				write(fd, mcptr, nel * sizeof(mem_cache));
				close(fd);
			}
			
			strncpy(mckey.domain, host, 255);
			mckey.domain[255] = '\0';
			
			mcval = (mem_cache*) bsearch(&mckey, mcptr, nel, sizeof(mem_cache), mccmp);
			
			if (mcval) {
				char *rval = apr_pstrdup(r->pool, mcval->path);
				
				env = apr_pstrdup(r->pool, mcval->extra);
				
				free(mcptr);
				
				mysql_free_result(res);
				mysql_close(mysql);
				
				return rval;
			}
			
			mysql_free_result(res);
			mysql_close(mysql);
			
			free(mcptr);
			
			return NULL;
		}
		else {
			int fd;
			void *pa;
			mem_cache *mcptr, mckey, *mcval;
			
			if ((fd = open(v2hr->id[9], O_RDONLY, 0)) == -1) {
				ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h: path_lookup (read img phase): unable to open: %s: %s.", v2hr->id[9], strerror(errno));
				
				v2hr->fl &= ~f_image;
				
				goto mysql_lookup;
			}
			
			pa = mmap((caddr_t) 0, st.st_size, PROT_READ, MAP_SHARED, fd, 0);
			mcptr = (mem_cache*) pa;
			
			strncpy(mckey.domain, host, 255);
			mckey.domain[255] = '\0';
			
			mcval = (mem_cache*) bsearch(&mckey, mcptr, st.st_size / sizeof(mem_cache), sizeof(mem_cache), mccmp);
			
			if (mcval) {
				char *pptr = apr_pstrdup(r->pool, mcval->path);
				
				env = apr_pstrdup(r->pool, mcval->extra);
				
				munmap((char*) pa, st.st_size);
				close(fd);
				return pptr;
			}
			
			munmap((char*) pa, st.st_size);
			close(fd);
			return NULL;
		}
	}
	else {
		mysql_lookup:
		return mysql_host_check(r, host, env, v2hr);
	}
}

static int read_post(request_rec *r, apr_table_t **tab)
{
	int rc;
	const char *type, *val, *key;
	char *rbuf = NULL;
	
	type = apr_table_get(r->headers_in, "Content-Type");
	
	if (strcasecmp(type, "application/x-www-form-urlencoded")) {
		return DECLINED;
	}
	
	if ((rc = ap_setup_client_block(r, REQUEST_CHUNKED_ERROR)) != OK) {
		return rc;
	}
	
	if (ap_should_client_block(r)) {
		char buf[HUGE_STRING_LEN];
		int rsize, len_read, rpos = 0;
		long length = r->remaining;
		
		rbuf = (char*) apr_pcalloc(r->pool, length + 1);

		while ((len_read = ap_get_client_block(r, buf, sizeof(buf))) > 0) {
			
			if ((rpos + len_read) > length) {
				rsize = length - rpos;
			}
			else {
				rsize = len_read;
			}
			
			memcpy(rbuf + rpos, buf, rsize);
			rpos += rsize;
		}
	}
	
	if (*tab) {
		apr_table_clear(*tab);
	}
	else {
		*tab = apr_table_make(r->pool, 2);
	}
	
	if (rbuf) {
		while (*rbuf && (val = ap_getword(r->pool, (const char**) &rbuf, '&'))) {
			key = ap_getword(r->pool, &val, '=');
			
			ap_unescape_url((char*) key);
			ap_unescape_url((char*) val);
			
			apr_table_merge(*tab, key, val);
		}
	}
	
	return OK;
}

static int v2h_init_handler(apr_pool_t *pconf, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s)
{
	ap_add_version_component(pconf, "v2h/2.0.4");
	
	return OK;
}

static int v2h_translate(request_rec *r)
{
	v2h_config_rec *v2hr = (v2h_config_rec*) ap_get_module_config(r->server->module_config, &v2h_module);
	const char *host;
	char *path, *env = NULL;
	
	if (r->uri[0] != '/') {
		return DECLINED;
	}
	
	host = apr_table_get(r->headers_in, "Host");
	
	if (!host) {
		ap_log_error(APLOG_MARK, APLOG_ERR|APLOG_NOERRNO, 0, r->server, "v2h_handler error: no host found: %s", host);
		
		return DECLINED;
	}
	
	path = path_lookup(r, host, env, v2hr);
	
	if (!path) {
		return DECLINED;
	}
	
	apr_table_set(r->subprocess_env, "V2H_HOST", host);
	apr_table_set(r->subprocess_env, "V2H_PATH", path);
	apr_table_set(r->subprocess_env, "V2H_EXTRA", env ? env : "");
	
	for (plist<char*>::node *n = v2hr->decline_list->top; n; n = n->next) {
		if (match(n->e, r->uri)) {
			return DECLINED;
		}
	}
	
	r->filename = apr_psprintf(r->pool, "%s%s%s", v2hr->id[8] ? v2hr->id[8] : "", path, r->uri);
	
	return OK;
}

static int v2h_handler(request_rec *r)
{
	v2h_config_rec *v2hr = (v2h_config_rec*) ap_get_module_config(r->server->module_config, &v2h_module);
	apr_table_t *tab = NULL;
	
	if (strcmp("v2h-handler", r->handler)) {
		return DECLINED;
	}
	
	r->content_type = "text/html";
	
	if (r->method_number == M_POST) {
		int rc;
		const char *ok;
		
		if ((rc = read_post(r, &tab)) != OK) {
			return rc;
		}
		
		ok = apr_table_get(tab, "submit");
		
		if (!strcmp(ok, "_clear_memory")) {
			unlink(v2hr->id[9]);
		}
		
		ap_rputs("<HTML><BODY BACKGROUND=\"#E0E0E0\"><H1>Memory Cleared</H1></BODY></HTML>", r);
		
		return OK;
	}
	
	ap_rprintf(r, "<HTML><BODY BACKGROUND=\"E0E0E0\"><CENTER><FORM METHOD=POST ACTION=%s><INPUT TYPE=submit VALUE=\"Clear VHE Memory\"><INPUT TYPE=hidden NAME=submit VALUE=\"_clear_memory\"></FORM></CENTER></BODY></HTML>", r->uri);
	
	return OK;
}
